/* csv2xlsx: make an Office Open XML file from files of  comma-separated values. */
/* Original date: 2024-10-10 */
/* Version: 1.1 */
/* Date: 2024-10-31 */

#include <ctype.h>
#include <errno.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "xlsxwriter.h"
#define DEFAULT_OUTPUT_FILE_NAME "output.xlsx"

/* Read a CSV file and add its values to a new sheet. */
static bool
read_file (const char *const program_name,
	   const char *const file_name,
	   const char *const sheet_name, lxw_workbook * const workbook)
{
  bool return_value = true;
  const int verbosity = 0;
  FILE *fp = stdin;
  if (strcmp (file_name, "-") == 0)
    {
      if (verbosity > 1)
	{
	  puts ("read standard input");
	}
    }
  else
    {
      fp = fopen (file_name, "r");
      if (fp == NULL)
	{
	  fprintf (stderr, "%s: ", program_name);
	  perror (file_name);
	  return_value = false;
	}
      else if (verbosity > 1)
	{
	  printf ("read input file: %s\n", file_name);
	}
    }
  if (fp != NULL)
    {
      size_t sheet_array_size = 1;
      lxw_worksheet **sheet_array =
	malloc (sheet_array_size * sizeof (*sheet_array));
      if (sheet_array == NULL)
	{
	  sheet_array_size = 0;
	  perror (program_name);
	  return_value = false;
	}
      else
	{
	  size_t sheet = 0;
	  if (sheet_name != NULL)
	    {
	      const lxw_error error =
		workbook_validate_sheet_name (workbook, sheet_name);
	      if (error != LXW_NO_ERROR)
		{
		  fprintf (stderr,
			   "%s: %s: invalid sheet name '%s': liblxlsxwriter error %d: %s\n",
			   program_name, file_name, sheet_name, error,
			   lxw_strerror (error));
		  return_value = false;
		}
	      else
		{
		  sheet_array[sheet] =
		    workbook_add_worksheet (workbook, sheet_name);
		}
	    }
	  if (sheet_array[sheet] == NULL && fp != stdin)
	    {
	      const char *s = strrchr (file_name, '/');
	      if (s == NULL)
		{
		  s = file_name;
		}
	      else
		{
		  s++;
		}
	      size_t len = strlen (s);
	      if (len >= 4 && strcmp (s + len - 4, ".csv") == 0)
		{
		  len -= 4;
		}
	      const size_t size = len + 1;
	      char *str = malloc (size);
	      if (str == NULL)
		{
		  perror (program_name);
		  return_value = false;
		}
	      else
		{
		  strncpy (str, s, len);
		  str[len] = '\0';
		  const lxw_error error =
		    workbook_validate_sheet_name (workbook, str);
		  if (error == LXW_NO_ERROR)
		    {
		      sheet_array[sheet] =
			workbook_add_worksheet (workbook, str);
		    }
		  free (str);
		  str = NULL;
		}
	    }
	  if (sheet_array[sheet] == NULL)
	    {
	      sheet_array[sheet] = workbook_add_worksheet (workbook, NULL);
	    }
	  if (sheet_array[sheet] == NULL)
	    {
	      fprintf (stderr, "%s: %s: cannot add sheet\n", program_name,
		       file_name);
	      return_value = false;
	    }
	  else
	    {
	      if (verbosity > 2)
		{
		  printf ("sheet name: %s\n", sheet_array[sheet]->name);
		}
	      size_t size = 1;
	      char *str = malloc (size);
	      if (str == NULL)
		{
		  perror (program_name);
		  return_value = false;
		}
	      else
		{
		  size_t len = 0;
		  int c;
		  unsigned long long int line = 1;
		  bool carriage_return = false;
		  unsigned long long int start_line = 0;
		  lxw_col_t column = 0;
		  lxw_row_t row = 0;
		  bool quoted = false;
		  bool quote = false;
		  do
		    {
		      c = fgetc (fp);
		      if (c == '\n')
			{
			  line++;
			  if (carriage_return == true && len != 0)
			    {
			      len--;
			      carriage_return = false;
			    }
			}
		      else if (c == '\r')
			{
			  carriage_return = true;
			}
		      else if (carriage_return == true)
			{
			  carriage_return = false;
			}
		      if (quoted == true)
			{
			  if (c == '"')
			    {
			      if (quote == true)
				{
				  quote = false;
				  continue;
				}
			      else
				{
				  quote = true;
				}
			    }
			  else if (quote == true)
			    {
			      len--;
			      quoted = false;
			      quote = false;
			    }
			}
		      if (len == 0 && quoted == false && c == '"')
			{
			  start_line = line;
			  quoted = true;
			}
		      else if (quoted == false
			       && (c == '\n' || c == ',' || c == EOF))
			{
			  if (len != 0)
			    {
			      str[len] = '\0';
			      if (verbosity > 4)
				{
				  printf ("%" PRIu32 ", %" PRIu16 ": %s\n",
					  row, column, str);
				}
			      const lxw_error error =
				worksheet_write_string (sheet_array[sheet],
							row, column, str,
							NULL);
			      if (error != LXW_NO_ERROR)
				{
				  fprintf (stderr,
					   "%s: %s:%llu: sheet %s, row %"
					   PRIu32 ", column %" PRIu16
					   ": liblxlsxwriter error %d: %s\n",
					   program_name, file_name, line,
					   sheet_array[sheet]->quoted_name,
					   row, column, error,
					   lxw_strerror (error));
				}
			      len = 0;
			    }
			  if (c == '\n')
			    {
			      sheet = 0;
			      column = 0;
			      row++;
			      if (row == 1 << 20)
				{
				  sheet_array[sheet] =
				    workbook_add_worksheet (workbook, NULL);
				  if (sheet_array[sheet] == NULL)
				    {
				      fprintf (stderr,
					       "%s: %s:%lluu: cannot add sheet\n",
					       program_name, file_name, line);
				      return_value = false;
				      break;
				    }
				  else
				    {
				      if (verbosity > 2)
					{
					  printf ("sheet name: %s\n",
						  sheet_array[sheet]->name);
					}
				      for (sheet = 1;
					   sheet < sheet_array_size; sheet++)
					{
					  sheet_array[sheet] = NULL;
					}
				      sheet = 0;
				      row = 0;
				    }
				}
			    }
			  else if (c == ',')
			    {
			      column++;
			      if (column == 1 << 14)
				{
				  sheet++;
				  if (sheet == sheet_array_size)
				    {
				      lxw_worksheet **const new_sheet_array =
					realloc (sheet_array,
						 ++sheet_array_size *
						 sizeof (*sheet_array));
				      if (new_sheet_array == NULL)
					{
					  sheet_array_size--;
					  perror (program_name);
					  return_value = false;
					  break;
					}
				      else
					{
					  sheet_array = new_sheet_array;
					  sheet_array[sheet] = NULL;
					}
				    }
				  if (sheet_array[sheet] == NULL)
				    {
				      sheet_array[sheet] =
					workbook_add_worksheet (workbook,
								NULL);
				      if (sheet_array[sheet] == NULL)
					{
					  fprintf (stderr,
						   "%s: %s:%lluu: cannot add sheet\n",
						   program_name, file_name,
						   line);
					  return_value = false;
					  break;
					}
				      else
					{
					  if (verbosity > 2)
					    {
					      printf ("sheet name: %s\n",
						      sheet_array[sheet]->
						      name);
					    }
					}
				    }
				  column = 0;
				}
			    }
			}
		      else
			{
			  if (len == size - 1)
			    {
			      size = size * 1.1 + 1;
			      char *const new_str = realloc (str, size);
			      if (new_str == NULL)
				{
				  str[len] = '\0';
				  perror (program_name);
				  return_value = false;
				  break;
				}
			      else
				{
				  str = new_str;
				}
			    }
			  str[len++] = c;
			}
		    }
		  while (c != EOF);
		  if (feof (fp) == 0 && ferror (fp) != 0)
		    {
		      fprintf (stderr, "%s: ", program_name);
		      perror (file_name);
		      return_value = false;
		    }
		  else if (quoted == true && c == EOF)
		    {
		      str[len] = '\0';
		      fprintf (stderr, "%s:%llu: unclosed quoted entry: %s\n",
			       file_name, start_line, str);
		      return_value = false;
		    }
		  free (str);
		  str = NULL;
		}
	    }
	  if (fp != stdin)
	    {
	      if (0 != fclose (fp))
		{
		  fprintf (stderr, "%s: ", program_name);
		  perror (file_name);
		  return_value = false;
		}
	    }
	  free (sheet_array);
	  sheet_array = NULL;
	  sheet_array_size = 0;
	}
    }
  return return_value;
}

int
main (int argc, char *argv[])
{
  const char *const program_name = argv[0];
  int return_value = EXIT_SUCCESS;
  enum
  { output_file_name, input_file_name, last_string_opt };
  const char *string_opt[last_string_opt] = { NULL };
  bool end = false;
  unsigned long int sheet_name_array_size = 0;
  unsigned long int sheet_name_count = 0;
  const char **sheet_name_array = NULL;
  unsigned long int file_count = 0;
  const int verbosity = 0;
  char *derived_output_file_name = NULL;
  lxw_workbook *workbook = NULL;
  if (argc > 1)
    {
      enum opt_type
      { info_opt_type, string_opt_type, string_array_opt_type };
      enum
      { help, version, last_info_opt };
      struct opt
      {
	const char *const short_opt;
	const char *const long_opt;
	const char *const help;
	unsigned char len;
	const int type;
	const int item;
      };
      struct opt opt_array[] = {
	{"o FILE", "output-file-name=FILE",
	 "Set the output file name to FILE.\nThe default output file name copies the first input file name\n(without any .csv file extension) and adds a .xlsx file extension.\nIf there is no input file name,\nthe default output file name is "
	 DEFAULT_OUTPUT_FILE_NAME ".", 0, string_opt_type, output_file_name},
	{"s NAME", "sheet-name=NAME",
	 "Add NAME to the list of sheet names.\nThe nth input file uses the nth sheet name.\nIf the nth sheet name is absent or invalid,\nthe sheet name copies the input file name\n(without any directory or .csv file extension).\nIf this is not a valid sheet name, the sheet uses a generic sheet name.\nIf an input file has more rows or colums than can fit on a sheet,\nthe entries overflow to new sheets with generic names.",
	 0, string_array_opt_type, 0},
	{"h", "help", "Output this help and exit.", 0, info_opt_type, help},
	{"V", "version", "Output version information and exit.", 0,
	 info_opt_type, version},
      };
      const size_t opt_count = sizeof (opt_array) / sizeof (opt_array[0]);
      size_t j;
      for (j = 0; j < opt_count; j++)
	{
	  const int type = opt_array[j].type;
	  const int item = opt_array[j].item;
	  if (type == info_opt_type && item < last_info_opt)
	    {
	    }
	  else if (type == string_opt_type && item < last_string_opt)
	    {
	    }
	  else if (type == string_array_opt_type)
	    {
	    }
	  else
	    {
	      fprintf (stderr,
		       "%s: internal error: invalid option: %zu (type %d, item %d)\n",
		       program_name, j, type, item);
	      return_value = EXIT_FAILURE;
	      end = true;
	    }
	  if (opt_array[j].long_opt != NULL)
	    {
	      const char *const s = strchr (opt_array[j].long_opt, '=');
	      if (s != NULL)
		{
		  opt_array[j].len = s - opt_array[j].long_opt;
		}
	      else
		{
		  opt_array[j].len = strlen (opt_array[j].long_opt);
		}
	    }
	}
      unsigned int pass;
      for (pass = 0; pass < 3 && end == false; pass++)
	{
	  bool opt_end = false;
	  int i;
	  for (i = 1; i < argc && end == false; i++)
	    {
	      if (opt_end == false && argv[i][0] == '-' && argv[i][1] != '\0')
		{
		  if (argv[i][1] == '-' && argv[i][2] == '\0')
		    {
		      opt_end = true;
		    }
		  else
		    {
		      size_t opt_match_count = 0;
		      const char *s = argv[i] + 1;
		      const char *v = NULL;
		      bool more = false;
		      do
			{
			  size_t opt_match = 0;
			  if (more == false && *s == '-')
			    {
			      s++;
			      size_t j;
			      size_t len = strlen (s);
			      v = strchr (s, '=');
			      if (v != NULL)
				{
				  len = v - s;
				  v++;
				}
			      for (j = 0; j < opt_count; j++)
				{
				  if (opt_array[j].long_opt != NULL)
				    {
				      if (v == NULL
					  || opt_array[j].
					  long_opt[opt_array[j].len] == '=')
					{
					  if (len == opt_array[j].len
					      && strncmp (s,
							  opt_array[j].
							  long_opt, len) == 0)
					    {
					      opt_match_count = 1;
					      opt_match = j;
					      break;
					    }
					  else if (len < opt_array[j].len
						   && strncmp (s,
							       opt_array[j].
							       long_opt,
							       len) == 0)
					    {
					      if (opt_match_count == 0)
						{
						  opt_match_count = 1;
						  opt_match = j;
						}
					      else
						if ((opt_array[j].type !=
						     opt_array[opt_match].
						     type)
						    || (opt_array[j].item !=
							opt_array[opt_match].
							item))
						{
						  opt_match_count++;
						}
					    }
					}
				    }
				}
			      if (opt_match_count == 0)
				{
				  fprintf (stderr,
					   "%s: unrecognized option '%s'\n",
					   program_name, argv[i]);
				  return_value = EXIT_FAILURE;
				  end = true;
				}
			      else if (opt_match_count == 1)
				{
				  if (v == NULL)
				    {
				      if (opt_array[opt_match].
					  long_opt[opt_array[opt_match].
						   len] == '=')
					{
					  if (i + 1 < argc)
					    {
					      v = argv[++i];
					    }
					  else
					    {
					      fprintf (stderr,
						       "%s: option --%.*s must have an argument\n",
						       program_name,
						       opt_array[opt_match].
						       len,
						       opt_array[opt_match].
						       long_opt);
					      return_value = EXIT_FAILURE;
					      end = true;
					    }
					}
				    }
				}
			      else
				{
				  size_t j;
				  fprintf (stderr,
					   "%s: option '%s' is ambiguous: '--%.*s'",
					   program_name, argv[i],
					   opt_array[opt_match].len,
					   opt_array[opt_match].long_opt);
				  for (j = opt_match + 1; j < opt_count; j++)
				    {
				      if (opt_array[j].long_opt != NULL)
					{
					  if (v == NULL
					      || opt_array[j].
					      long_opt[opt_array[j].len] ==
					      '=')
					    {
					      if (len < opt_array[j].len
						  && strncmp (s,
							      opt_array[j].
							      long_opt,
							      len) == 0)
						{
						  fprintf (stderr,
							   ", '--%.*s'",
							   opt_array[j].len,
							   opt_array[j].
							   long_opt);
						}
					    }
					}
				    }
				  fputc ('\n', stderr);
				  return_value = EXIT_FAILURE;
				  end = true;
				}
			    }
			  if (end == false)
			    {
			      if (opt_match_count == 0)
				{
				  const unsigned c = *s++;
				  size_t j;
				  more = false;
				  for (j = 0; j < opt_count; j++)
				    {
				      if (opt_array[j].short_opt != NULL)
					{
					  if (c == opt_array[j].short_opt[0])
					    {
					      opt_match = j;
					      if (opt_array[opt_match].
						  short_opt[1] == ' ')
						{
						  if (*s != '\0')
						    {
						      v = s;
						      s = strchr (s, '\0');
						    }
						  else if (i + 1 < argc)
						    {
						      v = argv[++i];
						    }
						  else
						    {
						      fprintf (stderr,
							       "%s: option %c must have an argument\n",
							       program_name,
							       c);
						      return_value =
							EXIT_FAILURE;
						      end = true;
						    }
						}
					      else if (*s != '\0')
						{
						  more = true;
						}
					      break;
					    }
					}
				    }
				  if (j == opt_count)
				    {
				      if (isprint (c) != 0)
					{
					  fprintf (stderr,
						   "%s: invalid option '%c'\n",
						   program_name, c);
					}
				      else
					{
					  fprintf (stderr,
						   "%s: invalid option '\\x%02X'\n",
						   program_name, c);
					}
				      return_value = EXIT_FAILURE;
				      end = true;
				    }
				}
			      if (end == false)
				{
				  const int type = opt_array[opt_match].type;
				  const int item = opt_array[opt_match].item;
				  if (pass == 0)
				    {
				      if (type == info_opt_type)
					{
					  if (item == help)
					    {
					      size_t j;
					      printf
						("Usage: %s [OPTIONS] [FILES]\n",
						 program_name);
					      puts
						("Read one or more comma-separated values (CSV) files and");
					      puts
						("use the libxlsxwriter library to make a Office Open XML spreadsheet file\n(usually with the .xlsx file extension).");
					      puts
						("Each CSV file uses a separate sheet in the output file.");
					      puts
						("With no file names or with the file name -, read standard input.\n");
					      for (j = 0; j < opt_count; j++)
						{
						  if (opt_array[j].
						      short_opt != NULL
						      && opt_array[j].
						      long_opt != NULL)
						    {
						      printf ("-%s, --%s\n",
							      opt_array[j].
							      short_opt,
							      opt_array[j].
							      long_opt);
						    }
						  else if (opt_array[j].
							   short_opt != NULL)
						    {
						      printf ("-%s\n",
							      opt_array[j].
							      short_opt);
						    }
						  else if (opt_array[j].
							   long_opt != NULL)
						    {
						      printf ("--%s\n",
							      opt_array[j].
							      long_opt);
						    }
						  if (opt_array[j].help !=
						      NULL)
						    {
						      puts (opt_array[j].
							    help);
						    }
						}
					      puts ("\nExamples:");
					      puts
						("\nMake data.xlsx with a sheet with the name data from data.csv:");
					      printf ("%s data.csv\n",
						      program_name);
					      printf
						("%s -o data.xlsx -s data data.csv\n",
						 program_name);
					      printf
						("%s --output-file-name=data.xlsx --sheet-name=data data.csv\n",
						 program_name);
					      printf
						("cat data.csv | %s -o data.xlsx -s data\n",
						 program_name);
					      puts
						("\nMake data.xlsx with sheets with the names data1, data2, and data3 from the files file1.csv, file2.csv, and file3.csv:");
					      printf
						("%s -o data.xlsx -s data1 -s data2 -s data3 file1.csv file2.csv file3.csv\n",
						 program_name);
					    }
					  else if (item == version)
					    {
					      puts ("csv2xlsx 1.1");
					      printf ("libxlsxwriter %s\n",
						      lxw_version ());
					    }
					  end = true;
					}
				      else if (type == string_opt_type)
					{
					  if (*v == '\0')
					    {
					      if (opt_match_count == 1)
						{
						  fprintf (stderr,
							   "%s: option --%.*s must not have a blank argument\n",
							   program_name,
							   opt_array
							   [opt_match].len,
							   opt_array
							   [opt_match].
							   long_opt);
						}
					      else
						{
						  fprintf (stderr,
							   "%s: option %c must not have a blank argument\n",
							   program_name,
							   opt_array
							   [opt_match].
							   short_opt[0]);
						}
					      return_value = EXIT_FAILURE;
					      end = true;
					    }
					  else
					    {
					      string_opt[item] = v;
					    }
					}
				      else if (type == string_array_opt_type)
					{
					  sheet_name_count++;
					}
				    }
				  else if (pass == 1)
				    {
				      if (type == string_array_opt_type)
					{
					  if (sheet_name_count <
					      sheet_name_array_size)
					    {
					      sheet_name_array
						[sheet_name_count++] = v;
					    }
					}
				    }
				}
			    }
			}
		      while (more == true && end == false);
		    }
		}
	      else if (pass == 0)
		{
		  file_count++;
		  if (string_opt[output_file_name] == NULL
		      && string_opt[input_file_name] == NULL)
		    {
		      if (argv[i][0] != '-' || argv[i][1] != '\0')
			{
			  string_opt[input_file_name] = argv[i];
			}
		    }
		}
	      else if (pass == 2)
		{
		  const char *sheet_name = NULL;
		  if (file_count < sheet_name_array_size)
		    {
		      sheet_name = sheet_name_array[file_count];
		    }
		  file_count++;
		  if (read_file (program_name, argv[i], sheet_name, workbook)
		      == false)
		    {
		      return_value = EXIT_FAILURE;
		    }
		}
	    }
	  if (pass == 0 && end == false)
	    {
	      if (string_opt[output_file_name] == NULL)
		{
		  if (string_opt[input_file_name] != NULL)
		    {
		      size_t len = strlen (string_opt[input_file_name]);
		      if (len >= 4
			  && strcmp (string_opt[input_file_name] + len - 4,
				     ".csv") == 0)
			{
			  len -= 4;
			}
		      const size_t size = len + 6;
		      derived_output_file_name = malloc (size);
		      if (derived_output_file_name == NULL)
			{
			  perror (program_name);
			  return_value = EXIT_FAILURE;
			  end = true;
			}
		      else
			{
			  strncpy (derived_output_file_name,
				   string_opt[input_file_name], len);
			  strcpy (derived_output_file_name + len, ".xlsx");
			  if (verbosity != 0)
			    {
			      printf ("derived output file name: %s\n",
				      derived_output_file_name);
			    }
			  string_opt[output_file_name] =
			    derived_output_file_name;
			}
		    }
		  else
		    {
		      string_opt[output_file_name] = DEFAULT_OUTPUT_FILE_NAME;
		    }
		}
	      if (end == false)
		{
		  if (sheet_name_count != 0)
		    {
		      if (file_count == 0)
			{
			  sheet_name_array_size = 1;
			}
		      else if (sheet_name_count <= file_count)
			{
			  sheet_name_array_size = sheet_name_count;
			}
		      else
			{
			  sheet_name_array_size = file_count;
			}
		      sheet_name_array =
			malloc (sheet_name_array_size *
				sizeof (*sheet_name_array));
		      if (sheet_name_array == NULL)
			{
			  sheet_name_array_size = 0;
			  perror (program_name);
			  return_value = EXIT_FAILURE;
			  end = true;
			}
		      else
			{
			  sheet_name_count = 0;
			  file_count = 0;
			}
		    }
		}
	      if (end == false)
		{
		  workbook = workbook_new (string_opt[output_file_name]);
		  if (workbook == NULL)
		    {
		      fprintf (stderr,
			       "%s: %s: libxlsxwriter cannot make workbook\n",
			       program_name, string_opt[output_file_name]);
		      return_value = EXIT_FAILURE;
		      end = true;
		    }
		}
	    }
	}
    }
  if (end == false)
    {
      if (file_count == 0)
	{
	  if (workbook == NULL)
	    {
	      if (string_opt[output_file_name] == NULL)
		{
		  string_opt[output_file_name] = DEFAULT_OUTPUT_FILE_NAME;
		}
	      workbook = workbook_new (string_opt[output_file_name]);
	      if (workbook == NULL)
		{
		  fprintf (stderr,
			   "%s: %s: libxlsxwriter cannot make workbook\n",
			   program_name, string_opt[output_file_name]);
		  return_value = EXIT_FAILURE;
		  end = true;
		}
	    }
	  if (end == false)
	    {
	      const char *sheet_name = NULL;
	      if (sheet_name_array_size != 0)
		{
		  sheet_name = sheet_name_array[file_count];
		}
	      if (read_file (program_name, "-", sheet_name, workbook) ==
		  false)
		{
		  return_value = EXIT_FAILURE;
		}
	    }
	}
    }
  if (workbook != NULL)
    {
      const lxw_error error = workbook_close (workbook);
      if (error != LXW_NO_ERROR)
	{
	  fprintf (stderr, "%s: %s: liblxlsxwriter error %d: %s\n",
		   program_name, string_opt[output_file_name], error,
		   lxw_strerror (error));
	  return_value = EXIT_FAILURE;
	  end = true;
	}
      else
	{
	  if (verbosity != 0)
	    {
	      printf ("output file: %s\n", string_opt[output_file_name]);
	    }
	}
    }
  if (sheet_name_array != NULL)
    {
      free (sheet_name_array);
      sheet_name_array = NULL;
      sheet_name_array_size = 0;
    }
  if (derived_output_file_name != NULL)
    {
      free (derived_output_file_name);
      derived_output_file_name = NULL;
    }
  if (0 != fclose (stdout))
    {
      perror (program_name);
      return_value = EXIT_FAILURE;
    }
  return return_value;
}
